from .equations import *
from numpy import arange
from math import degrees
from .well import Well, define_section


def get(mdt, profile='V', build_angle=1, kop=0, eob=0, sod=0, eod=0, kop2=0, eob2=0, **kwargs):
    """
    Generate a wellpath.

    Parameters
    ----------
    mdt: num
        target depth, m or ft
    profile: str
        'V' for vertical, 'J' for J-type, 'S' for S-type, 'H1' for Horizontal single curve and 'H2' for Horizontal double curve
    build_angle: num
        building angle, °
    kop: num
        kick-off point, m or ft
    eob: num
        end of build, m or ft
    sod: num
        start of drop, m or ft
    eod: num
        end of drop, m or ft
    kop2: num
        kick-off point 2, m or ft
    eob2: num
        end of build 2, m or ft

    Keyword Args
    ------------
    points: int
        number of points
    set_start: dict, None
        set initial point in m {'north': 0, 'east': 0}.
    change_azimuth: float, int, None
        add specific degrees to azimuth values along the entire well.
    set_info: dict, None
        dict, {'dlsResolution', 'wellType': 'onshore'|'offshore', 'units': 'metric'|'english'}.

    Returns
    -------
    well: well object
        A wellpath object with 3D position
    """

    # Settings
    params = {'points': 100, 'set_start': None, 'change_azimuth': None, 'set_info': None, 'ndigits': 2}
    for key, value in kwargs.items():
        params[key] = value
    set_start = params['set_start']
    change_azimuth = params['change_azimuth']
    set_info = params['set_info']

    info = {'dlsResolution': 30, 'wellType': 'offshore', 'units': 'metric'}

    initial_point = {'north': 0, 'east': 0, 'depth': 0}

    if isinstance(set_info, dict):
        for param in set_info:  # changing default values
            if param in info:
                info[param] = set_info[param]

    if isinstance(set_start, dict):
        for x in set_start:  # changing default values
            if x in initial_point:
                initial_point[x] = set_start[x]

    md = list(arange(0, mdt + 1, 1))    # Measured Depth from RKB, m
    depth_step = md[1]

    if profile == 'V':        # Vertical well
        tvd, north, east, inc, az = vertical_section(profile, md, kop, depth_step)

    elif profile == 'J':        # J-type well
        tvd, north, east, inc, az = create_j_well(mdt, md, kop, eob, build_angle, depth_step)

    elif profile == 'S':  # S-type well
        tvd, north, east, inc, az = create_s_well(mdt, md, kop, eob, sod, eod, build_angle, depth_step)

    elif profile == 'H1':     # Horizontal single-curve well
        tvd, north, east, inc, az = create_h1_well(mdt, md, kop, eob, depth_step)

    else:        # Horizontal double-curve well
        tvd, north, east, inc, az = create_h2_well(mdt, md, kop, eob, kop2, eob2, build_angle, depth_step)

    if change_azimuth is not None:
        az, north_new, east_new = mod_azimuth(change_azimuth, az, north, east)

    # CREATING TRAJECTORY POINTS
    trajectory = [{'md': 0, 'inc': 0, 'azi': 0, 'dl': 0, 'tvd': 0, 'sectionType': 'vertical'}]
    trajectory[-1].update(initial_point)
    for idx, md in enumerate(md):
        if idx > 0:
            dogleg = calc_dogleg(inc[idx - 1], inc[idx], az[idx - 1], az[idx])
            point = {'md': md, 'inc': inc[idx], 'azi': az[idx],
                     'north': calc_north(trajectory[-1]['north'], trajectory[-1]['md'], md,
                                         trajectory[-1]['inc'], inc[idx],
                                         trajectory[-1]['azi'], az[idx],
                                         dogleg),
                     'east': calc_east(trajectory[-1]['east'], trajectory[-1]['md'], md,
                                       trajectory[-1]['inc'], inc[idx],
                                       trajectory[-1]['azi'], az[idx],
                                       dogleg),
                     'tvd': calc_tvd(trajectory[-1]['tvd'], trajectory[-1]['md'], md,
                                     trajectory[-1]['inc'], inc[idx], dogleg),
                     'dl': degrees(dogleg)
                     }
            point['sectionType'] = define_section(point, trajectory[-1])
            trajectory.append(point)

    return Well({'trajectory': trajectory, 'info': info})


def vertical_section(profile, md, kop, depth_step):
    if profile == 'V':
        tvd = md
    else:
        tvd = md[:round(kop / depth_step) + 1]  # True Vertical Depth from RKB, m

    north = [0] * len(tvd)  # x axis
    east = [0] * len(tvd)  # x axis
    inclination = [0.0] * len(tvd)
    azimuth = [0] * len(tvd)

    return tvd, north, east, inclination, azimuth


def create_s_well(mdt, md, kop, eob, sod, eod, build_angle, depth_step):
    # Vertical section
    tvd, north, east, inclination, azimuth = vertical_section('S', md, kop, depth_step)

    # Build section
    s = depth_step
    theta_delta = radians(build_angle) / round((eob - kop) / depth_step)
    theta = theta_delta
    r = s / theta

    z_displacement = (r * sin(theta))
    tvd.append(round(tvd[-1] + z_displacement, 2))
    z_count = z_displacement

    hz_displacement = r * (1 - cos(theta))
    north.append(round(north[-1] + hz_displacement, 2))
    east.append(0)
    inclination.append(degrees(theta))
    azimuth.append(0)

    for x in range(round((eob - kop) / depth_step) - 1):
        theta += theta_delta
        inclination.append(degrees(theta))
        z_displacement = (r * sin(theta)) - z_count
        tvd.append(round(tvd[-1] + z_displacement, 2))
        z_count += z_displacement

        hz_displacement = r * (1 - cos(theta)) - north[-1]
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        azimuth.append(0)

    # Tangent section
    z_displacement = (depth_step * cos(radians(build_angle)))
    hz_displacement = (depth_step * sin(radians(build_angle)))

    for x in range(round((sod - eob) / depth_step)):
        tvd.append(round(tvd[-1] + z_displacement, 2))
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        inclination.append(inclination[-1])
        azimuth.append(0)

    # Drop section
    s = depth_step
    points_drop = round((eod - sod) / depth_step)
    theta_delta = radians(build_angle) / points_drop
    theta = radians(build_angle)
    r = s / theta_delta
    z_checkpoint = tvd[-1]
    hz_checkpoint = north[-1]
    for x in range(points_drop):
        z_displacement = r * (sin(theta) - sin(theta - (theta_delta * (x + 1))))
        tvd.append(round(z_checkpoint + z_displacement, 2))

        hz_displacement = r * (1 - cos(theta)) - r * (1 - cos(theta - (theta_delta * (x + 1))))
        north.append(round(hz_checkpoint + hz_displacement, 2))
        east.append(0)
        inclination.append(inclination[-1] - degrees(theta_delta))
        azimuth.append(0)

    # Vertical section
    for x in range(round((mdt - eod) / depth_step)):
        tvd.append(round(tvd[-1] + depth_step, 2))
        north.append(north[-1])  # x axis
        east.append(0)
        inclination.append(0)
        azimuth.append(0)

    return tvd, north, east, inclination, azimuth


def create_j_well(mdt, md, kop, eob, build_angle, depth_step):
    # Vertical section
    tvd, north, east, inclination, azimuth = vertical_section('J', md, kop, depth_step)

    # Build section
    s = depth_step
    theta_delta = radians(build_angle / round((eob - kop) / depth_step))
    theta = theta_delta
    r = s / theta

    z_vertical = tvd[-1]
    z_displacement = (r * sin(theta))
    tvd.append(round(tvd[-1] + z_displacement, 2))

    hz_displacement = r * (1 - cos(theta))
    north.append(round(north[-1] + hz_displacement, 2))
    east.append(0)
    inclination.append(degrees(theta))
    azimuth.append(0)

    for x in range(round((eob - kop) / depth_step) - 1):
        theta += theta_delta
        inclination.append(degrees(theta))

        z_displacement = (r * sin(theta))
        tvd.append(round(z_vertical + z_displacement, 2))

        hz_displacement = r * (1 - cos(theta)) - north[-1]
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        azimuth.append(0)

    # Tangent section
    z_displacement = (depth_step * cos(radians(build_angle)))
    hz_displacement = (depth_step * sin(radians(build_angle)))
    for x in range(round((mdt - eob) / depth_step)):
        tvd.append(round(tvd[-1] + z_displacement, 2))
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        inclination.append(inclination[-1])
        azimuth.append(0)

    return tvd, north, east, inclination, azimuth


def create_h1_well(mdt, md, kop, eob, depth_step):
    # Vertical section
    tvd, north, east, inclination, azimuth = vertical_section('H1', md, kop, depth_step)

    # Build section
    s = depth_step
    theta_delta = radians(90) / round((eob - kop) / depth_step)
    theta = theta_delta
    r = s / theta

    z_displacement = (r * sin(theta))
    tvd.append(round(tvd[-1] + z_displacement, 2))
    z_count = z_displacement

    hz_displacement = r * (1 - cos(theta))
    north.append(round(north[-1] + hz_displacement, 2))
    east.append(0)
    inclination.append(degrees(theta))
    azimuth.append(0)

    for x in range(round((eob - kop) / depth_step) - 1):
        theta += theta_delta
        z_displacement = (r * sin(theta)) - z_count
        tvd.append(round(tvd[-1] + z_displacement, 2))
        z_count += z_displacement

        hz_displacement = r * (1 - cos(theta)) - north[-1]
        inclination.append(degrees(theta))
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        azimuth.append(0)

    # Horizontal section
    for x in range(round((mdt - eob) / depth_step)):
        tvd.append(tvd[-1])
        north.append(north[-1] + depth_step)
        east.append(0)
        inclination.append(90)
        azimuth.append(0)

    return tvd, north, east, inclination, azimuth


def create_h2_well(mdt, md, kop, eob, kop2, eob2, build_angle, depth_step):
    # Vertical section
    tvd, north, east, inclination, azimuth = vertical_section('H2', md, kop, depth_step)

    # Build section
    s = depth_step
    theta_delta = radians(build_angle / round((eob - kop) / depth_step))
    theta = theta_delta
    r = s / theta

    z_displacement = (r * sin(theta))
    tvd.append(round(tvd[-1] + z_displacement, 2))
    z_count = z_displacement

    hz_displacement = r * (1 - cos(theta))
    north.append(round(north[-1] + hz_displacement, 2))
    east.append(0)
    inclination.append(degrees(theta))
    azimuth.append(0)

    for x in range(round((eob - kop) / depth_step) - 1):
        theta = theta + theta_delta
        z_displacement = (r * sin(theta)) - z_count
        tvd.append(round(tvd[-1] + z_displacement, 2))
        z_count += z_displacement

        hz_displacement = r * (1 - cos(theta)) - north[-1]
        inclination.append(degrees(theta))
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        azimuth.append(0)

    # Tangent section
    z_displacement = (depth_step * cos(radians(build_angle)))
    hz_displacement = (depth_step * sin(radians(build_angle)))
    for x in range(round((kop2 - eob) / depth_step)):
        tvd.append(round(tvd[-1] + z_displacement, 2))
        inclination.append(inclination[-1])
        north.append(round(north[-1] + hz_displacement, 2))
        east.append(0)
        azimuth.append(0)

    # Build section 2
    s = depth_step
    build_angle = 90 - build_angle
    points_drop = round((eob2 - kop2) / depth_step)
    theta_delta = radians(build_angle) / points_drop
    theta = radians(build_angle)
    r = s / theta_delta
    z_checkpoint = tvd[-1]
    hz_checkpoint = north[-1]

    for x in range(points_drop):
        hz_displacement = r * (sin(theta) - sin(theta - (theta_delta * (x + 1))))
        north.append(round(hz_checkpoint + hz_displacement, 2))
        inclination.append(inclination[-1] + degrees(theta_delta))
        east.append(0)
        azimuth.append(0)

        z_displacement = r * (1 - cos(theta)) - r * (1 - cos(theta - (theta_delta * (x + 1))))
        tvd.append(round(z_checkpoint + z_displacement, 2))

    # Horizontal section
    for x in range(round((mdt - eob2) / depth_step)):
        tvd.append(tvd[-1])
        north.append(north[-1] + depth_step)
        inclination.append(inclination[-1])
        east.append(0)
        azimuth.append(0)

    return tvd, north, east, inclination, azimuth


def mod_azimuth(change_azimuth, azimuth_new, north_new, east_new):
    for a in range(len(azimuth_new)):
        azimuth_new[a] += change_azimuth

        if change_azimuth <= 90:
            east_new[a] = north_new[a] * sin(radians(change_azimuth))
            north_new[a] *= cos(radians(change_azimuth))
        elif 90 < change_azimuth <= 180:
            angle = change_azimuth - 90
            east_new[a] = north_new[a] * round(cos(radians(angle)), 3)
            north_new[a] *= - sin(radians(angle))
        elif 180 < change_azimuth <= 270:
            angle = change_azimuth - 180
            east_new[a] = - north_new[a] * round(sin(radians(angle)), 3)
            north_new[a] *= - cos(radians(angle))
        else:
            angle = change_azimuth - 270
            east_new[a] = - north_new[a] * round(cos(radians(angle)), 3)
            north_new[a] *= sin(radians(angle))

    return azimuth_new, north_new, east_new



from .equations import *
from .plot import plot_wellpath, plot_top_view, plot_vs
import pandas as pd


class Well(object):
    def __init__(self, data):
        self.info = data['info']
        self.trajectory = data['trajectory']
        for idx, point in enumerate(self.trajectory):
            if idx > 0:
                delta_md = point['md'] - self.trajectory[idx - 1]['md']
                point['dls'] = calc_dls(point, delta_md, resolution=self.info['dlsResolution'])
                point['delta'] = get_delta(point, self.trajectory[idx-1])
            else:
                point['dls'] = 0
                point['delta'] = get_delta(point)
        self.npoints = len(self.trajectory)

    def plot(self, **kwargs):
        default = {'plot_type': '3d', 'add_well': None, 'names': None, 'style': None, 'y_axis': 'md', 'x_axis': 'inc'}
        for key, value in kwargs.items():
            default[key] = value

        if default['plot_type'] == '3d':
            fig = plot_wellpath(self, add_well=default['add_well'], names=default['names'], style=default['style'])
            return fig
        elif default['plot_type'] == 'top':
            fig = plot_top_view(self, add_well=default['add_well'], names=default['names'], style=default['style'])
            return fig
        elif default['plot_type'] == 'vs':
            fig = plot_vs(self, y_axis=default['y_axis'], x_axis=default['x_axis'], add_well=default['add_well'],
                          names=default['names'], style=default['style'])
            return fig
        else:
            raise ValueError('The plot type "{}" is not recognised'.format(default['plot_type']))

    def df(self):
        dataframe = pd.DataFrame(self.trajectory)
        return dataframe

    def add_location(self, lat, lon):
        """
        set specific location lat and lon in decimal degrees
        """
        self.info['location'] = {'lat': lat, 'lon': lon}

    def add_reference(self, references):
        self.info['rkb'] = references['rkb']
        if 'waterDepth' in references:
            self.info.update({'waterDepth': references['waterDepth'],
                              'seabed': references['rkb'] + references['waterDepth']})


def define_section(p2, p1=None):

    if not p1:
        return 'vertical'

    else:
        if p2['inc'] == p1['inc'] == 0:
            return 'vertical'
        else:
            if round(p2['inc'], 2) == round(p1['inc'], 2):
                if p2['tvd'] == p1['tvd'] == 0:
                    return 'horizontal'  # Horizontal Section
                else:
                    return 'hold'  # Straight Inclined Section
            else:
                if p2['inc'] > p1['inc']:
                    return 'build-up'   # Built-up Section
                if p2['inc'] < p1['inc']:
                    return 'drop-off'   # Drop-off Section


def get_delta(p2, p1=None):

    if not p1:
        return {'md': 0, 'tvd': 0, 'inc': 0, 'azi': 0, 'dl': 0, 'dls': 0, 'north': 0, 'east': 0}

    else:
        delta = {}
        for param in ['md', 'tvd', 'inc', 'azi', 'dl', 'dls', 'north', 'east']:
            delta.update({param: p2[param] - p1[param]})

    return delta



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot_wellpath(well, **kwargs):
    """
    Plot a 3D Wellpath.

    Keyword Arguments:
        well: a well object with 3D position,
        add_well: include a new well or list of wells
        names: set name or list of names for wells included in the plot
        style: {'darkMode': bool, # activate dark mode. default = False
                'color': str, # color by specific property. e.g. 'dls'|'dl'|'tvd'|'md'|'inc'|'azi'. default = None
                'size': num, # marker size. default = 2
                }

    Returns:
        3D Plot - plotly.graph_objects.Figure
    """

    data = {'add_well': None, 'names': None, 'style': None}
    for key, value in kwargs.items():
        data[key] = value

    units = well.info['units']

    well1 = pd.DataFrame(well.trajectory)
    well1["well"] = 1
    result = well1

    if data['add_well'] is not None:
        wells = []

        if type(data['add_well']) is not list:
            data['add_well'] = [data['add_well']]

        well_no = 2
        for x in data['add_well']:
            new_well = pd.DataFrame(x.trajectory)
            new_well["well"] = well_no
            wells.append(new_well)
            well_no += 1

        all_wells = well1.append(wells)
        result = all_wells

    if data['names'] is not None:

        if type(data['names']) is not list:
            data['names'] = [data['names']]

        well_no = 1
        for x in data['names']:
            result.replace({'well': {well_no: x}}, inplace=True)
            well_no += 1

    style = define_style(data['style'])

    if style['color'] is None:
        color = 'well'
        fig = px.line_3d(result, x="east", y="north", z="tvd", color=color)

    else:
        fig = go.Figure(
            data=[go.Scatter3d(x=result['east'],
                               y=result['north'],
                               z=result['tvd'],
                               mode='markers',
                               marker=dict(
                                    size=style['size'],
                                    color=result[style['color']],  # set color to an array/list of desired values
                                    showscale=True,
                                    opacity=0.8),
                               legendgroup=True,
                               hovertemplate='%{text}<extra></extra><br>' + '<b>North</b>: %{y:.2f}<br>' +
                                             '<b>East</b>: %{x}<br>' + '<b>TVD</b>: %{z}<br>',
                               text=result['well'])])

    if units == 'metric':
        fig.update_layout(scene=dict(
            xaxis_title='East, m',
            yaxis_title='North, m',
            zaxis_title='TVD, m',
            aspectmode='manual'))
    else:
        fig.update_layout(scene=dict(
            xaxis_title='East, ft',
            yaxis_title='North, ft',
            zaxis_title='TVD, ft',
            aspectmode='manual'))
    fig.update_scenes(zaxis_autorange="reversed")
    fig.layout.template = style['darkMode']
    include_logo(fig, style['darkMode'], plot_type='3D', color=style['color'])
    fig.update_layout(title='Wellbore Trajectory - 3D View')

    return fig


def plot_top_view(well, **kwargs):
    data = {'add_well': None, 'names': None, 'style': None}
    for key, value in kwargs.items():
        data[key] = value

    wells = [well]
    units = well.info['units']

    if data['add_well'] is not None:
        if type(data['add_well']) is not list:
            data['add_well'] = [data['add_well']]
        wells += data['add_well']

    if data['names']:
        if type(data['names']) is not list:
            data['names'] = [data['names']]
    else:
        data['names'] = []
        for idx in range(len(wells)):
            data['names'].append('well ' + str(idx + 1))

    fig = go.Figure()

    for idx, w in enumerate(wells):
        fig.add_trace(go.Scatter(
            x=[point['east'] for point in w.trajectory],
            y=[point['north'] for point in w.trajectory],
            hovertemplate='<b>North</b>: %{y:.2f}<br>' + '<b>East</b>: %{x}<br>',
            showlegend=False, name=data['names'][idx]))

    if units == 'metric':
        fig.update_layout(xaxis_title='East, m',
                          yaxis_title='North, m')
    else:
        fig.update_layout(xaxis_title='East, ft',
                          yaxis_title='North, ft')

    fig.update_layout(title='Wellbore Trajectory - Top View', hovermode='closest')

    style = define_style(data['style'])
    include_logo(fig, style['darkMode'], plot_type='top')
    fig.layout.template = style['darkMode']

    return fig


def plot_vs(well, **kwargs):
    unit_system = well.info['units']
    dls_res = well.info['dlsResolution']
    data = {'y_axis': 'md', 'x_axis': 'inc', 'add_well': None, 'names': None, 'style': None}
    for key, value in kwargs.items():
        data[key] = value

    possible_props = ['md', 'tvd', 'inc', 'azi', 'dl', 'dls']
    for prop in [data['x_axis'], data['y_axis']]:
        if prop not in possible_props:
            raise ValueError('The axis "{}" is not recognised'.format(prop))

    wells = [well]

    if data['add_well'] is not None:
        if type(data['add_well']) is not list:
            data['add_well'] = [data['add_well']]
        wells += data['add_well']

    if data['names']:
        if type(data['names']) is not list:
            data['names'] = [data['names']]
    else:
        data['names'] = []
        for idx in range(len(wells)):
            data['names'].append('well ' + str(idx+1))

    fig = go.Figure()

    for idx, w in enumerate(wells):
        fig.add_trace(go.Scatter(
            x=[point[data['x_axis']] for point in w.trajectory],
            y=[point[data['y_axis']] for point in w.trajectory],
            hovertemplate='<b>y</b>: %{y:.2f}<br>' + '<b>x</b>: %{x:.2f}<br>',
            showlegend=False, name=data['names'][idx]))

    units = ['m', '°']
    for key, axis, in {'0': data['x_axis'], '1': data['y_axis']}.items():
        if axis in ['md', 'tvd']:
            if unit_system == 'metric':
                units[int(key)] = 'm'
            else:
                units[int(key)] = 'ft'
        elif axis in ['inc', 'azi', 'dl']:
            units[int(key)] = '°'
        elif axis == 'dls':
            if unit_system == 'metric':
                units[int(key)] = '°/' + str(dls_res) + 'm'
            else:
                units[int(key)] = '°/' + str(dls_res) + 'ft'

    style = define_style(data['style'])
    include_logo(fig, style['darkMode'], plot_type='vs')
    fig.layout.template = style['darkMode']

    fig.update_layout(xaxis_title=data['x_axis'] + ', ' + units[0],
                      yaxis_title=data['y_axis'] + ', ' + units[1],
                      title='Wellbore Trajectory - ' + data['x_axis'] + ' vs ' + data['y_axis'],
                      hovermode='closest')

    return fig


def define_style(style):
    set_style = {'darkMode': False, 'color': None, 'size': 2}
    if style is not None:
        for key in style.keys():
            set_style[key] = style[key]

    if set_style['darkMode']:
        set_style['darkMode'] = 'plotly_dark'
    else:
        set_style['darkMode'] = None

    return set_style


def include_logo(fig, dark=None, plot_type='3D', color=None):
    if not color:
        if not dark:
            y = 0.91
        else:
            y = 0.95
    else:
        if not dark:
            y = 0.97
        else:
            y = 1
    if dark == 'plotly_dark':
        if plot_type != '3D':
            y = 1.03
        fig.update_layout(images=[dict(
                          source="https://raw.githubusercontent.com/jcamiloangarita/opensource_apps/master/"
                                 "resources/pwp-logo-dark.png",
                          xref="paper", yref="paper",
                          x=0.5, y=y,
                          sizex=0.12, sizey=0.12,
                          xanchor="right", yanchor="bottom")])
    else:
        if plot_type != '3D':
            y = 0.99
        fig.update_layout(images=[dict(
                          source="https://avatars.githubusercontent.com/u/53001276?v=4",
                          xref="paper", yref="paper",
                          x=0.35, y=y,
                          sizex=0.2, sizey=0.2,
                          xanchor="right", yanchor="bottom")])



from math import atan, degrees, radians, asin
from numpy import linspace
import pandas as pd
from .load_trajectory import load


def two_points(points, inner_points=20):
    """
    Arguments:
        points: {'kickoff':{'north': num, 'east': num, 'tvd': num},
                 'target': {'north': num, 'east': num, 'tvd': num}}
        inner_points: number of points between curved zone

    Returns:
        a wellpath object with 3D position
    """

    if 'north' not in points['kickoff']:
        points['kickoff']['north'] = 0

    if 'east' not in points['kickoff']:
        points['kickoff']['east'] = 0

    point_1 = points['kickoff']
    point_2 = points['target']

    # set first section
    trajectory = [{'md': 0, 'inc': 0, 'azi': 0},
                  {'md': point_1['tvd'], 'inc': 0, 'azi': 0}]

    # calculate deltas
    delta = {'vertical': point_2['tvd'] - point_1['tvd'],
             'north': point_2['north'] - point_1['north'],
             'east': point_2['east'] - point_1['east']}

    delta['horizontal'] = (delta['north']**2 + delta['east']**2)**0.5

    # Define azimuth
    azimuth = 0
    if delta['north'] != 0 and delta['east'] != 0:
        beta = degrees(atan(delta['north'] / delta['east']))
        if delta['east'] > 0:
            azimuth = 90 - beta
        else:
            azimuth = 270 - beta

    else:
        if delta['north'] == 0:
            if delta['east'] > 0:
                azimuth = 90
            else:
                azimuth = 270
        if delta['east'] == 0:
            if delta['north'] > 0:
                azimuth = 0
            else:
                azimuth = 180

    # 3 cases comparing vertical and horizontal displacement
    steps = inner_points + 1
    if delta['vertical'] == delta['horizontal']:
        radius = delta['horizontal']
        theta = 90
        arc = radius * radians(theta)

        new_md = linspace(point_1['tvd']+arc/steps, point_1['tvd']+arc, steps)
        new_inc = linspace(theta/steps, theta, steps)

        for md, inc in zip(new_md, new_inc):
            trajectory.append({'md': md, 'inc': inc, 'azi': azimuth})

        well = load(pd.DataFrame(trajectory), equidistant=False, set_start=point_1)

        return well

    if delta['vertical'] < delta['horizontal']:
        # curve section
        radius = delta['vertical']
        theta = 90
        arc = radius * radians(theta)

        new_md = linspace(point_1['tvd'] + arc / steps, point_1['tvd'] + arc, steps)
        new_inc = linspace(theta / steps, theta, steps)
        for md, inc in zip(new_md, new_inc):
            trajectory.append({'md': md, 'inc': inc, 'azi': azimuth})

        # horizontal section
        trajectory.append({'md': trajectory[-1]['md']+(delta['horizontal']-delta['vertical']), 'inc': 90,
                           'azi': trajectory[-1]['azi']})

        well = load(pd.DataFrame(trajectory), equidistant=False, set_start=point_1)

        return well

    if delta['vertical'] > delta['horizontal']:
        if delta['horizontal'] != 0:
            radius = (delta['horizontal']**2 + delta['vertical']**2)/(2*delta['horizontal'])
            theta = degrees(asin(delta['vertical']/radius))
            arc = radius * radians(theta)
            new_md = linspace(point_1['tvd'] + arc / steps, point_1['tvd'] + arc, steps)
            new_inc = linspace(theta / steps, theta, steps)
        else:
            new_md = [point_2['tvd']]
            new_inc = [0]
        for md, inc in zip(new_md, new_inc):
            trajectory.append({'md': md, 'inc': inc, 'azi': azimuth})

        well = load(pd.DataFrame(trajectory), equidistant=False, set_start=point_1)

        return well
